from __future__ import annotations

from pathlib import Path
from typing import Any, Union

import yaml

from pls.args import args
from pls.exceptions import ConfigException
from pls.models.node_spec import NodeSpec


def load_yaml_file(file_path: Path) -> Any:
    """
    Load the YAML file referenced by the given name.

    :param file_path: the path to the YAML file to read and parse
    :return: the parsed contents of the YAML file
    """

    with file_path.open("r") as data_file:
        data = yaml.safe_load(data_file)
    return data


def internal_yml_path(file_name: str) -> Path:
    """
    Map the given YAML data file name to the path, assuming that all the data
    files are in the same directory as the file of this function.

    :param file_name: the name of the YAML data file
    :return: the path to the YAML data file
    """

    return Path(__file__).parent.joinpath(file_name)


def massage_specs(entry: dict) -> list[dict]:
    """
    For convenience, specs can be written as group combining several names,
    patterns or extensions. This function splits such groups into its
    constituent specs.
    Grouping does not allow mixing names, patterns and extensions.

    :param entry: a single entry from the node specs
    :return: a list of specs split from entry
    """

    id_fields = ["name", "pattern", "extension"]
    singular_plural_map = {field: f"{field}s" for field in id_fields}
    all_id_fields = [*singular_plural_map.keys(), *singular_plural_map.values()]

    # Exactly one identification method should be present.
    if [field in entry for field in all_id_fields].count(True) != 1:
        methods = ", ".join([f"`{method}`" for method in all_id_fields])
        raise ConfigException(f"Exactly one of {methods} is required.")

    # Split plurals if present.
    for singular, plural in singular_plural_map.items():
        if plural in entry:
            if not isinstance(entry[plural], list):
                raise ConfigException(
                    f"`{plural}` must be a list. Use `{singular}` instead."
                )
            common = {k: v for k, v in entry.items() if k != plural}
            return [{singular: value, **common} for value in entry[plural]]

    # Ensure no singular is list.
    for singular, plural in singular_plural_map.items():
        if singular in entry and isinstance(entry[singular], list):
            raise ConfigException(
                f"`{singular}` cannot be a list. Use `{plural}` instead."
            )

    return [entry]


def parse_node_specs(specs: list[dict]) -> list[NodeSpec]:
    """
    Parse information about the list of all node specs for all languages. This
    maps the POPO representation of the specs into the ``NodeSpec`` instances.

    :param specs: the list of POPO specs as read from the spec definition files
    :return: the list of all node specs for all languages
    """

    return [NodeSpec(**spec) for entry in specs for spec in massage_specs(entry)]


def locate_config() -> Union[Path, None]:
    """
    Find a config file with the name ``.pls.yml`` in the directory or its
    ancestors, upto a max depth based on CLI arguments.

    :return: the path to the file if found, ``None`` otherwise
    """

    config_name = ".pls.yml"
    curr_dir: Path = args.directory
    for i in range(args.depth):
        config_path = curr_dir.joinpath(config_name)
        if config_path.exists() and config_path.is_file():
            return config_path
        curr_dir = curr_dir.parent
    return None


conf_data = load_yaml_file(conf_path) if (conf_path := locate_config()) else {}

node_specs: list[NodeSpec] = parse_node_specs(
    load_yaml_file(internal_yml_path("node_specs.yml"))
)
"""a list of all node specs for all languages, read from ``node_specs.yml``"""
if node_specs_ext := conf_data.get("node_specs"):
    node_specs = parse_node_specs(node_specs_ext) + node_specs

nerd_icons: dict[str, str] = load_yaml_file(internal_yml_path("nerd_icons.yml"))
"""a mapping of icon names to Unicode code-points, read from ``nerd_icons.yml``"""
if nerd_icons_ext := conf_data.get("nerd_icons"):
    nerd_icons.update(nerd_icons_ext)

emoji_icons: dict[str, str] = load_yaml_file(internal_yml_path("emoji_icons.yml"))
"""a mapping of icon names to emoji, read from ``emoji_icons.yml``"""
if emoji_icons_ext := conf_data.get("emoji_icons"):
    emoji_icons.update(emoji_icons_ext)



from __future__ import annotations

import multiprocessing
import os
from pathlib import Path
from typing import Union

from rich.console import Console

from pls.args import args
from pls.enums.node_type import NodeType
from pls.enums.sort_order import SortOrder
from pls.models.node import Node
from pls.state import State, state


console = Console()


def sort_key(node: Node) -> str:
    """
    Map a ``Node`` instance to a string that represents it. This string is used
    to sort a list of ``Node`` instances.

    :param node: the node item to reduce to a string
    :return: a string representative of the node
    """

    key = node.name.lstrip(".").lower()
    if not args.no_dirs_first:
        prefix = "0" if node.node_type == NodeType.DIR else "1"
        key = f"{prefix}{key}"
    return key


def parse_nodes(node_name: str, parent_state: State) -> Union[Node, None]:
    """
    Parse the node name into a ``Node`` instance. Most of the heavy lifting is
    handled in the ``Node`` class definition itself.

    This function is executed by multiprocessing workers and thus must be
    passed the state explicitly from the main process.

    :param node_name: the name of a node inside the working directory
    :param parent_state: the global state of the parent application
    :return: a ``Node`` instance
    """

    node_path: Path = args.directory.joinpath(node_name)

    if node_path.is_dir():
        if args.no_dirs:
            return None
    else:  # is some kind of file
        if args.no_files:
            return None

    return Node(node_name, path=node_path, state=parent_state)


def read_input() -> list[Node]:
    """
    Get a list of all directories and files in the given directory.

    :return: the list of directories and files inside the given directory
    """

    all_nodes = os.listdir(args.directory)

    if not all_nodes:
        console.print(
            f"There are no files or folders in [bold]{args.directory}[/bold].",
            highlight=False,
        )
    else:
        with multiprocessing.Pool() as pool:
            comp_nodes = pool.starmap(
                parse_nodes,
                [(node, state) for node in all_nodes],
            )
        all_nodes = [node for node in comp_nodes if node is not None]
        all_nodes.sort(key=sort_key, reverse=args.sort == SortOrder.DESC)

    return all_nodes



from __future__ import annotations

from rich.console import Console
from rich.table import Table

from pls.args import args
from pls.enums.icon_type import IconType
from pls.models.node import Node
from pls.state import state


console = Console()

column_spec = {
    "": {"name": ""},  # dummy column to act like spacer
    "type": {
        # 'type' is a pseudo-column linked to 'perms', so it has no name.
        "name": ""
    },
    "perms": {"name": "Permissions"},
    "user": {"name": "User"},
    "group": {"name": "Group"},
    "size": {"name": "Size", "attrs": {"justify": "right"}},
    "git": {"name": "Git"},
    "icon": {
        # 'icon' is a pseudo-column linked to 'name', so it has no name.
        "name": "",
        "attrs": {"width": 2},
    },
    "name": {
        # The names have a leading space when the leading dots are aligned.
        "name": "Name"
        if args.no_align
        else " Name"
    },
}
"""a mapping of column keys to column spec"""

settings = {
    "padding": (0, 1, 0, 0),
    "box": None,
    "show_header": args.details,
    "header_style": "underline",
}
"""the settings for the Rich table"""


def get_columns() -> list[str]:
    """
    Get the list of columns to show.

    :return: the list of column keys
    """

    cols = []
    if args.details:
        cols.extend(["type", "perms", "", "user", "group", "", "size", ""])
    if args.details and state.is_git_managed:
        cols.extend(["git", ""])
    if args.icon != IconType.NONE:
        cols.append("icon")
    cols.append("name")
    return cols


def get_table() -> Table:
    """
    Get a Rich table with pre-configured columns. The attributes of the columns
    are retrieved from ``column_spec`` based on keys from ``get_columns``.

    :return: a Rich table
    """

    table = Table(**settings)
    for col_key in get_columns():
        col = column_spec.get(col_key)
        table.add_column(col.get("name"), **col.get("attrs", {}))
    return table


def write_output(all_nodes: list[Node]):
    """
    Write the list of directories and files to the screen as a table.

    :param all_nodes: the list of directories and files
    """

    table = get_table()

    for node in all_nodes:
        data = node.table_row
        if data is not None:
            cells = [data.get(col, "") for col in get_columns()]
            table.add_row(*cells)

    console.print(table)



import argparse
import os
from pathlib import Path

from pls import __version__
from pls.enums.icon_type import IconType
from pls.enums.sort_order import SortOrder
from pls.enums.unit_system import UnitSystem
from pls.exceptions import ExecException


parser = argparse.ArgumentParser(
    prog="pls",
    description=(
        "`pls` is a better `ls` for developers. "
        "See https://github.com/dhruvkb/pls for more information."
    ),
)

########################
# Positional arguments #
########################


def directory(path: str) -> Path:
    """
    Parse the given path into a ``Path`` instance. Raise
    :param path: the path supplied as a CLI argument
    :return: the ``Path`` instance wrapping the supplied path
    :raise: ``ExecException``, if the path does not point to an actual directory
    """

    path = Path(path).resolve()
    if not os.path.isdir(path):
        raise ExecException("`directory` must be a path to a directory")
    return path


parser.add_argument(
    "directory",
    type=directory,
    nargs="?",  # makes the `directory` arg optional
    default=os.getcwd(),
    help="the directory whose contents are to be listed",
)

######################
# Optional arguments #
######################

parser.add_argument(
    *["-v", "--version"],
    action="version",
    version=f"%(prog)s {__version__}",
    help="show the version of the codebase",
)

parser.add_argument(
    *["-i", "--icon"],
    type=IconType,
    choices=list(IconType),
    default=IconType.NERD,
    help="the type of icons to show with the files",
)
parser.add_argument(
    *["-s", "--sort"],
    type=SortOrder,
    choices=list(SortOrder),
    default=SortOrder.ASC,
    help="the direction in which to sort the files and directories",
)
parser.add_argument(
    *["-d", "--depth"],
    type=int,
    default=4,
    help="the max depth upto which to look for a `.pls.yml` file",
)
parser.add_argument(
    *["-u", "--units"],
    type=UnitSystem,
    choices=list(UnitSystem),
    default=UnitSystem.BINARY,
    help="the units to use when listing the size of files",
)

parser.add_argument(
    "--details",
    action="store_true",
    help="show details such as permissions, owner and size",
)
parser.add_argument(
    "--all",
    action="store_true",
    help="show all files, including those that would otherwise be hidden",
)

parser.add_argument(
    "--no-dirs",
    action="store_true",
    help="hide directories in the output",
)
parser.add_argument(
    "--no-files",
    action="store_true",
    help="hide files in the output",
)
parser.add_argument(
    "--no-align",
    action="store_true",
    help="turn off character alignment for leading dots",
)
parser.add_argument(
    "--no-dirs-first",
    action="store_true",
    help="mix directories and files, sorting them together",
)


args = parser.parse_args()
"""the CLI arguments parsed by ``argparse``"""

__all__ = ["args"]



from enum import auto

from pls.enums.base import AutoEnum


class SortOrder(AutoEnum):
    """
    Nodes can be sorted in two orders:

    - ascending
    - descending

    This enum lists these possibilities.
    """

    ASC = auto()
    DESC = auto()



from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Union

from pls.exceptions import ExecException


def exec_git(cmd_args: list[str], cwd: Path) -> subprocess.CompletedProcess:
    """
    Execute a ``git`` command from the given directory.

    :param cmd_args: the args to pass in the Git command
    :param cwd: the directory in which to run the command
    :return: the ``CompletedProcess`` instance
    """

    if cmd_args is None:
        cmd_args = []
    cmd = ["git", *cmd_args]
    proc = subprocess.run(
        cmd,
        check=True,
        cwd=cwd,
        capture_output=True,
        text=True,
    )
    return proc


def get_git_root(working_dir: Path) -> Union[Path, None]:
    """
    Identify the Git root for the working directory. To get the root directory,
    this uses following command::

        git rev-parse --show-toplevel

    :return: the root dir ``Path`` if Git-managed, ``None`` otherwise
    """

    try:
        proc = exec_git(
            ["rev-parse", "--show-toplevel"],
            cwd=working_dir,
        )
        root_path = Path(proc.stdout.rstrip())
        return root_path
    except subprocess.CalledProcessError:
        return None


def get_git_statuses(git_root: Path) -> dict[Path, str]:
    """
    Identify the Git statuses for all files in the working directory. To get the
    Git statues, this uses following two commands::

        git status --porcelain --untracked-files --ignored
        git status --porcelain --untracked-files=normal --ignored=matching

    Refer to the `git-status command documentation
    <https://git-scm.com/docs/git-status>`_ for more info.

    :param git_root: the root directory of the Git repository
    :return: the mapping of paths to their Git statues
    """

    status_map: dict[Path, str] = {}

    status_lines: set[str] = set()
    try:
        status = ["status", "--porcelain"]

        proc = exec_git(
            [*status, "--untracked-files", "--ignored"],
            cwd=git_root,
        )
        status_lines.update(proc.stdout.rstrip().split("\n"))

        proc = exec_git(
            [*status, "--untracked-files=normal", "--ignored=matching"],
            cwd=git_root,
        )
        status_lines.update(proc.stdout.rstrip().split("\n"))
    except subprocess.CalledProcessError:
        return status_map

    for line in status_lines:
        status = line[0:2]

        components: list[str] = line[3:].split(" ")
        if len(components) == 1:
            path = components[0]
        elif len(components) == 3:
            _, __, path = components
        else:
            raise ExecException("Could not parse Git status code")
        path = Path(path)

        status_map[path] = status

    return status_map



from __future__ import annotations

from enum import auto

from pls.enums.base import AutoEnum


class UnitSystem(AutoEnum):
    """
    Bytes can be converted to higher units in two ways:

    - the decimal system where powers of 10 (factor: 10^3) are used
    - the binary system where powers of 2 (factor: 2^10) are used

    This enum lists these possibilities.
    """

    BINARY = auto()
    DECIMAL = auto()


def get_base_and_pad_and_units(us: UnitSystem) -> [int, list[str]]:
    """
    Get the base factor i.e. the ratio between any two successive orders of
    magnitude and the units corresponding to each multiplication of the base
    factor.

    :param us: the unit system in use
    :return: the base factor and the list of units
    """

    units = [
        "K",  # kibi | kilo
        "M",  # mebi | mega
        "G",  # gibi | giga
        "T",  # tebi | tera
        "P",  # pebi | peta
        "E",  # exbi | exa
        "Z",  # zebi | zetta
        "Y",  # yobi | yotta
    ]

    if us == UnitSystem.BINARY:
        base = pow(2, 10)
        pad = 3  # units will be 3 chars e.g. KiB, GiB
        units = ["", *map(lambda i: f"{i}i", units)]
    else:  # us == UnitSystem.DECIMAL:
        base = pow(10, 3)
        pad = 2  # units will be 2 chars e.g. KB, GB
        units = ["", *units]
    return base, pad, units



#!/usr/bin/env python3

from pls.data.getters import node_specs
from pls.fs.list import read_input
from pls.table import write_output


def main():
    """
    Represents the starting point of the application. This function:

    - accepts no inputs: options are read from CLI arguments using ``argparse``
    - returns no outputs: output is written to ``STDOUT`` using ``rich``
    """

    nodes = read_input()
    if not nodes:
        return
    for node in nodes:
        node.match(node_specs)
    write_output(nodes)


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from functools import cached_property
from pathlib import Path
from typing import Union

from pls.args import args
from pls.data.getters import emoji_icons, nerd_icons
from pls.enums.icon_type import IconType
from pls.enums.node_type import NodeType
from pls.fs.stats import (
    get_group,
    get_node_type,
    get_permission_text,
    get_size,
    get_user,
)
from pls.models.node_spec import NodeSpec
from pls.state import State


class Node:
    """
    A node is any file, folder or symlink on the file-system. This model stores
    attributes pertaining to a single FS node.

    Nodes are read from the file system directly using ``os.walk``.
    """

    def __init__(self, name: str, path: Path, state: State):
        self.name = name
        self.path = path

        self.state = state  # keeping a copy to pass to dest_nodes
        self.is_git_managed = state.is_git_managed
        if self.is_git_managed:
            self.path_wrt_git = path.relative_to(state.git_root)
            self.git_status = state.git_status_map.get(self.path_wrt_git, "  ")
        else:
            self.path_wrt_git = None
            self.git_status = None

        self.stat: Union[os.stat_result, None] = None
        try:
            self.stat = path.lstat()
            self.exists = True
        except FileNotFoundError:
            self.exists = False

        self.is_loop: bool = False  # only ``True`` for cyclic symlinks
        self.dest_node: Union[Node, str, None] = None  # only populated for symlinks

        self.specs: list[NodeSpec] = []  # matched later (see ``map_specs``)

    def __eq__(self, other: Node) -> bool:
        """
        Compare the object ``self`` to the other instance of ``Node``.

        :param other: the node to compare this node with for equality
        :return: ``True`` if the node instances are equal, ``False`` otherwise
        """

        return self.path.resolve() == other.path.resolve()

    def __repr__(self) -> str:
        """
        Get the string representation of the ``Node`` instance. This is also
        used by ``__str__`` automatically.

        :return: the string representation
        """

        return f"{self.name} @ {self.path}"

    @cached_property
    def node_type(self) -> NodeType:
        """whether the node is a file, folder, symlink, FIFO etc."""

        node_type = get_node_type(self.path)

        # Symlinks need to set their destination node.
        if node_type == NodeType.SYMLINK and self.dest_node is None:
            link_path = os.readlink(self.path)
            try:
                self.path.resolve()  # raises exception if cyclic

                # Use ``os.readlink`` instead of ``Path.resolve`` to step
                # through chained symlinks one-by-one.
                link = Path(link_path)
                if not link.is_absolute():
                    link = self.path.parent.joinpath(link)

                self.dest_node = Node(name=link_path, path=link, state=self.state)
            except RuntimeError as exc:
                if "Symlink loop" in str(exc):
                    self.is_loop = True
                    self.dest_node = link_path

        return node_type

    @cached_property
    def ext(self) -> Union[str, None]:
        """the extension of the node, i.e. the portion after the last dot"""

        return self.name.split(".")[-1] if "." in self.name else None

    @cached_property
    def format_pair(self) -> tuple[str, str]:
        """the opening and closing tags of Rich console formatting markup"""

        format_rules = []

        # font color
        if not self.exists:
            format_rules.append("red")  # only happens for broken symlinks
        elif spec_color := self.spec_attr("color"):
            format_rules.append(spec_color)
        elif self.node_type == NodeType.DIR:
            format_rules.append("cyan")

        # font weight
        if spec_importance := self.spec_attr("importance"):
            if spec_importance == 2:
                format_rules.append("underline")
            elif spec_importance == 1:
                format_rules.append("bold")
            elif spec_importance == -1:
                format_rules.append("dim")
        elif self.is_git_managed:
            if self.git_status == "!!":  # Git-ignored file
                format_rules.append("dim")

        # italics
        if self.name == ".pls.yml":
            format_rules.append("italic")

        if format_rules:
            left = f"[{' '.join(format_rules)}]"
            right = "[/]"
        else:
            left = right = ""
        return left, right

    @cached_property
    def formatted_suffix(self) -> str:
        """the symbol after the filename representing its type"""

        if not self.exists:
            return "⚠"

        if self.node_type == NodeType.SYMLINK:
            if self.is_loop:
                return f"[dim]@ ↺[/] [red]{self.dest_node}[/red]"
            return f"[dim]@ →[/] {self.dest_node.formatted_name}"

        mapping = {
            NodeType.DIR: "/",
            NodeType.SOCKET: "=",
            NodeType.FIFO: "|",
        }
        suffix = mapping.get(self.node_type, "")
        if suffix:
            suffix = f"[dim]{suffix}[/]"
        return suffix

    @cached_property
    def formatted_name(self) -> str:
        """the name, formatted using Rich console formatting markup"""

        name = self.name
        if self.formatted_suffix:
            name = f"{name}{self.formatted_suffix}"

        if name.startswith(".") and not args.no_align:
            name = name.replace(".", "[dim].[/dim]", 1)

        # Apply format pair.
        left, right = self.format_pair
        return f"{left}{name}{right}"

    @cached_property
    def formatted_icon(self) -> str:
        """the emoji or Nerd Font icon to show beside the node"""

        if args.icon == IconType.NONE:
            return ""

        if args.icon == IconType.EMOJI:
            icon_index = emoji_icons
        else:  # args.icon == IconType.NERD
            icon_index = nerd_icons

        if spec_icon := self.spec_attr("icon"):
            icon = icon_index.get(spec_icon)
        elif self.node_type == NodeType.DIR:
            icon = icon_index.get("folder")
        else:
            icon = None

        if icon:
            left, right = self.format_pair
            return f"{left}{icon}{right}"
        return ""

    @cached_property
    def formatted_git_status(self) -> str:
        """formatted two-letter Git status as returned by ``git-status``"""

        if self.git_status == "  ":
            return self.git_status

        format_map: dict[str, str] = {
            "M": "yellow",  # modified
            "A": "green",  # added
            "D": "red",  # deleted
            "!": "dim",  # ignored
            "-": "dim",  # padding
        }
        fmt_status = ""
        for letter in self.git_status:
            if letter == " ":
                letter = "-"
            if letter in format_map:
                fmt_status = f"{fmt_status}[{format_map[letter]}]{letter}[/]"
            else:
                fmt_status = f"{fmt_status}{letter}"
        return fmt_status

    @cached_property
    def is_visible(self) -> bool:
        """whether the node deserves to be rendered to the screen"""

        # If explicitly requested for all files, show all.
        if args.all:
            return True

        # Nodes without spec and with a leading dot are hidden.
        if not self.specs and self.name.startswith("."):
            return False

        # Nodes with importance -2 are hidden.
        if self.spec_attr("importance") == -2:
            return False

        return True

    @cached_property
    def type_char(self) -> str:
        """the single character representing the file type"""

        mapping = {
            NodeType.SYMLINK: "l",
            NodeType.DIR: "d",
            NodeType.FILE: "-",
            NodeType.FIFO: "p",
            NodeType.SOCKET: "s",
            NodeType.CHAR_DEVICE: "c",
            NodeType.BLOCK_DEVICE: "b",
        }
        return mapping[self.node_type]

    @cached_property
    def table_row(self) -> Union[dict[str, str], None]:
        """the mapping of column names and value when tabulating the node"""

        if not self.is_visible:
            return None

        cells = dict()

        name = self.formatted_name
        if not self.name.startswith(".") and not args.no_align:
            # Left pad name with a space to account for leading dots.
            name = f" {name}"
        cells["name"] = name

        cells["icon"] = self.formatted_icon

        if args.details:
            cells["type"] = self.type_char
            cells["perms"] = get_permission_text(self.stat.st_mode)
            cells["user"] = get_user(self.stat.st_uid)
            cells["group"] = get_group(self.stat.st_gid)
            if self.node_type != NodeType.DIR:
                cells["size"] = get_size(self.stat.st_size)

        if self.is_git_managed:
            cells["git"] = self.formatted_git_status

        return cells

    def spec_attr(self, attr: str) -> Union[str, int, None]:
        """
        Get the requested attribute from the first matching spec to provide it.

        :param attr: the requested attribute
        :return: the value of the attribute if found, ``None`` otherwise
        """

        for spec in self.specs:
            if attr_val := getattr(spec, attr, None):
                return attr_val
        return None

    def match(self, specs: list[NodeSpec]):
        """
        Find all spec matching this node from a list of all possible specs and
        store them in the ``specs`` attribute.

        :param specs: the list of all specs
        """

        self.specs = [spec for spec in specs if spec.match(self.name)]



from __future__ import annotations

from grp import getgrgid
from itertools import cycle
from pathlib import Path
from pwd import getpwuid

from pls.args import args
from pls.enums.node_type import NodeType, type_test_map
from pls.enums.unit_system import get_base_and_pad_and_units
from pls.exceptions import ExecException


def get_permission_text(st_mode: int) -> str:
    """
    Get the permission text for the node in the form of a triplet of 'rwx'
    strings.

    :param st_mode: the st_mode value of the ``os.stat_result`` instance
    :return: the text to render as the permissions of the node
    """

    perms = ["r", "w", "x"]
    color_map = {
        "r": "yellow",
        "w": "red",
        "x": "green",
        "T": "magenta",
        "s": "magenta",
    }
    perm_sets: list[list[str]] = [["-" for _ in range(3)] for _ in range(3)]

    text_rep = format(st_mode, "09b")[-9:]
    for index, (bit, perm) in enumerate(zip(text_rep, cycle(perms))):
        if int(bit):
            perm_sets[int(index / 3)][index % 3] = perm

    if st_mode & 0o4000 == 0o4000:  # setuid
        perm_sets[0][-1] = "s"
    if st_mode & 0o2000 == 0o2000:  # setgid
        perm_sets[1][-1] = "s"
    if st_mode & 0o1000 == 0o1000:  # sticky
        perm_sets[2][-1] = "T"

    return " ".join(
        "".join(
            f"[{color}]{perm}[/]" if (color := color_map.get(perm, "")) else perm
            for perm in perm_set
        )
        for perm_set in perm_sets
    )


def get_size(st_size: int) -> str:
    """
    Get the human-readable size of the node in the form of a number followed by
    a compound unit of a byte.

    :param st_size: the size of the node in bytes
    :return: the size of the node as a human-readable value
    """

    base, pad, units = get_base_and_pad_and_units(args.units)
    for index, unit in reversed(list(enumerate(units))):
        order_of_magnitude = base ** index
        if st_size >= order_of_magnitude:
            magnitude = round(st_size / order_of_magnitude)  # Pop! Pop!
            unit = f"{unit}B".rjust(pad, " ")

            return f"{magnitude}[dim]{unit}[/]"
    return f"{st_size}  [dim]B[/]"


def get_user(st_uid: int) -> str:
    """
    Get the name of the user that owns the node. This requires a ``passwd``
    lookup for the user ID found in the node stats.

    :param st_uid: the user ID mapped to the owner of the node
    :return: the name of the user who owns the node
    """

    return getpwuid(st_uid).pw_name


def get_group(st_gid: int) -> str:
    """
    Get the name of the group that owns the node. This requires a group database
    lookup for the group ID found in the node stats.

    :param st_gid: the group ID mapped to the owner of the node
    :return: the name of the group that owns the node
    """

    return getgrgid(st_gid).gr_name


def get_node_type(path: Path) -> NodeType:
    """
    Get the ``NodeType`` that corresponds to the given mode. This function uses
    functions defined in the ``stat`` module to identify the type.

    :param path: the path to the node
    :return: the right ``NodeType`` enum based on the mode
    """

    for node_type, node_type_test in type_test_map.items():
        if getattr(path, node_type_test)():
            return node_type
    else:
        raise ExecException("Could not determine type of the node.")



from __future__ import annotations

from enum import auto

from pls.enums.base import AutoEnum


class NodeType(AutoEnum):
    """
    A node can be either of these:

    - a directory
    - a regular file
    - a name FIFO pipe
    - a file-based socket
    - a char device
    - a block device
    - a symlink to any of the above

    This enum lists these possibilities. Refer to `the Wikipedia article on Unix
    file types <https://en.wikipedia.org/wiki/Unix_file_types>`_ for more info.
    """

    SYMLINK = auto()  # symbolic link
    DIR = auto()  # directory
    FILE = auto()  # regular file
    FIFO = auto()  # named pipe
    SOCKET = auto()  # socket
    CHAR_DEVICE = auto()  # character special device file
    BLOCK_DEVICE = auto()  # block special device file


type_test_map: dict[NodeType, str] = {
    node_type: f"is_{node_type.value}" for node_type in list(NodeType)
}
"""a mapping of node types with specific functions that evaluate it"""



from __future__ import annotations

from functools import cached_property
from pathlib import Path

from pls.args import args
from pls.fs.git import get_git_root, get_git_statuses


class State:
    """
    Holds the global state of the application.
    """

    def __init__(self):
        self.git_root = get_git_root(args.directory)
        self.git_status_map: dict[Path, str] = dict()
        if self.is_git_managed:
            self.git_status_map = get_git_statuses(self.git_root)

    def __repr__(self) -> str:
        """
        Get the string representation of the ``State`` instance. This is also
        used by ``__str__`` automatically.

        :return: the string representation
        """

        return str(self.__dict__)

    @cached_property
    def is_git_managed(self) -> bool:
        """whether the working directory is managed by Git"""

        return self.git_root is not None


state = State()
"""the global state of the application"""



from enum import Enum


class AutoEnum(Enum):
    def _generate_next_value_(self: str, *args) -> str:
        return self.lower()

    def __str__(self):
        """
        Get the string representation of this enum. Unlike other objects, this
        does not default to ``__repr__``.
        :return: the string representation
        """
        return self.value



from __future__ import annotations

import re


class NodeSpec:
    """
    A spec describes the nature of a class of FS nodes that determine its
    presentation. This model stores attributes pertaining to a single spec. Make
    sure to massage spec entries using :py:func:`pls.data.getters.massage_specs`
    before initialising.

    Node specs are read from ``node_spec.yml``.
    """

    def __init__(
        self,
        name: str = None,
        pattern: str = None,
        extension: str = None,
        icon: str = None,
        color: str = None,
        importance: int = 0,
    ):
        identification_methods = ["name", "pattern", "extension"]
        loc = locals()

        self.name = name
        self.pattern = re.compile(pattern) if pattern else None
        self.extension = extension

        self.icon = icon
        self.color = color
        self.importance = importance

    def __repr__(self) -> str:
        """
        Get the string representation of the ``NodeSpec`` instance. This is also
        used by ``__str__`` automatically.

        :return: the string representation
        """

        if self.name:
            return self.name
        if self.extension:
            return f"*.{self.extension}"
        if self.pattern:
            return f"<{self.pattern.pattern}>"

    def match(self, name: str) -> bool:
        """
        Check whether the given file name matches this spec.

        :param name: the name of the file to match against this spec
        :return: ``True`` if the file matches this entry, ``False`` otherwise
        """

        if self.name:
            return self.name == name
        elif self.pattern:
            return re.match(self.pattern, name) is not None
        elif self.extension:
            return name.endswith(f".{self.extension}")
        else:
            return False



from enum import auto

from pls.enums.base import AutoEnum


class IconType(AutoEnum):
    """
    The app supports three modes of icons. An icon can be either of these:

    - an emoji, which is supported by most modern OSes and terminals
    - a Nerd Font icon, which has more glyphs but needs separate installation
    - none, for plain output without fancy embellishments

    This enum lists these possibilities.
    """

    NERD = auto()
    EMOJI = auto()
    NONE = auto()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pls', 'pls.data', 'pls.enums', 'pls.fs', 'pls.models']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'rich>=11.1.0,<12.0.0']

entry_points = \
{'console_scripts': ['pls = pls.main:main']}

setup_kwargs = {
    'name': 'pls',
    'version': '1.8.1',
    'description': '`pls` is a better `ls` for developers.',
    'long_description': '<h1 align="center">\n  <img height="128px" src="https://raw.githubusercontent.com/dhruvkb/pls/main/readme_assets/pls.svg"/>\n</h1>\n\n<p align="center">\n  <a href="https://pypi.org/project/pls/">\n    <img src="https://img.shields.io/pypi/v/pls" alt="pls on PyPI"/>\n  </a>\n  <a href="https://www.python.org">\n    <img src="https://img.shields.io/pypi/pyversions/pls" alt="Python versions"/>\n  </a>\n  <a href="https://github.com/dhruvkb/pls/blob/main/LICENSE">\n    <img src="https://img.shields.io/github/license/dhruvkb/pls" alt="GPL-3.0"/>\n  </a>\n  <a href="https://github.com/dhruvkb/pls/actions/workflows/ci.yml">\n    <img src="https://github.com/dhruvkb/pls/actions/workflows/ci.yml/badge.svg" alt="CI status"/>\n  </a>\n</p>\n\n<p align="center">\n  <img src="https://raw.githubusercontent.com/dhruvkb/pls/main/readme_assets/demo.png" alt="Demo of `pls`"/>\n</p>\n\n`pls` is a better `ls` for developers. The "p" stands for "pro" (as in "professional"/"programmer") or "prettier", whichever helps you remember the command name.\n\nIt works in a manner similar to `ls`, in  that it lists directories and files in a given directory, but it adds many more [developer-friendly features](#features).\n\nNote that `pls` is not a replacement for `ls`. `ls` is a tried, tested and trusted command with lots of features. `pls`, on the other hand, is a simple tool for people who just want to see the contents of their directories.\n\n## Features\n\n`pls` provides many features over  `ls` command. `pls` can:\n\n- show Nerd Font icons or emoji next to files and directories making it easier to read the output\n- colour output to elevate important files or dim unimportant ones\n- use a more nuanced approach to hidden files than plainly hiding files with a leading dot `.`\n- group directories and shows them all before files\n- ignore leading dots `.` and normalise case when sorting files\n- align files names by first character\n- generate visibility rules by parsing `.gitignore`\n- show technical two-letter Git status for files and directories\n- cascade formatting rule specs by based on specificity levels\n- read [`.pls.yml`](.pls.yml) files from the directory to augment its configuration\n- show more details like permissions, owner and size in columns\n- run quite fast, thanks to multiprocessing\n\nThe icon, color and most behaviour in the application can be [configured using plain-text YAML files](src/pls/data/README.md) for the pros who prefer to tweak their tools.\n\n## Upcoming features\n\nIn the future `pls` will be able to\n\n- add MIME type as another method for matching files to specs\n- use complete path based matching for files\n- link files and hide derived files behind the main ones\n- generate tree-like output for subdirectories\n\nIf you want to help implement any of these features, feel free to submit a PR. `pls` is free and open-source software.\n\n## Comparison with similar tools\n\nThere are a lot of `ls` replacements. Here are some of the most popular ones.\n\n- [`exa`](https://github.com/ogham/exa)\n- [`lsd`](https://github.com/Peltoche/lsd)\n- [`colorls`](https://github.com/athityakumar/colorls)\n- [`ls-go`](https://github.com/acarl005/ls-go)\n\n`pls` aims to stand out because of some very specific choices.\n\n- Does not intend to replace `ls`. `pls`, as a command, is just as easy to type.\n- Targets a more tech-savvy audience in its [features](#features).\n- Intelligently [maps file type](src/pls/data/README.md). Just comparing the file extension would be too generic.\n- Meticulously chosen iconography for the appreciating eyes.\n- Highly customisable at a project level using a simple [`.pls.yml`](.pls.yml) file.\n- Built in a friendly language, Python. This makes it easy to fork and change it yourself.\n\n## Installation\n\nTo get the best of `pls`, [install a Nerd Font](https://github.com/ryanoasis/nerd-fonts/blob/master/readme.md#font-installation) on your computer. [Nerd Fonts](https://www.nerdfonts.com) come patched with many icons from different popular icon sets. If you\'re a "pro" (the target audience for `pls`) these fonts are basically a must.\n\n`pls` is a pure-Python codebase and is deployed to PyPI. So installing it on any system with a supported Python version is quite straightforward.\n\n```shell\n$ python3 -m pip install --user pls\n```\n\nThere are no native packages _yet_.\n\n## Usage\n\n`pls` has a very simple API with easy to memorise flags. There are no mandatory arguments. Just run `pls` anywhere on your disk.\n\n```shell\n$ pls\n```\n\nThere are a few optional arguments and flags you can use to tweak the behaviour. You can see the complete list of arguments and their description by passing the `--help` or `-h` flags.\n\n```shell\n$ pls --help\n```\n\n### Directory\n\nThe only positional argument is a directory. Pass this to see the contents of a different folder rather than the current working directory.\n\n```shell\n$ pls path/to/somewhere/else\n```\n\n### Icons\n\n`pls` supports many icons for popular languages out of the box and will show icons by default. If you don\'t have a Nerd Font (why?), you can switch to emoji icons using `--icons emoji` or `-iemoji`. Be warned they are quite bad. If you are a sad person, you turn icons off using `--icon none` or `-inone`.\n\n**Note:** The built-in icon configuration is intentionally lean. The whole idea is for `pls` to be [customisable by you](src/pls/data/README.md).\n\n### Filtering\n\nYou can choose to hide files or folders from the output using `--no-files` and `--no-dirs` respectively. Passing both will lead to a blank output. On the other hand if you want to see files and directories that `pls` would not show otherwise, pass `--all`.\n\n### Sorting\n\nBy default `pls` will place all directories first, followed by files with both sorted alphabetically from A to Z. You can prevent folders from being first by passing the `--no-dirs-first` flag. You can change the sort to go from Z to A using `--sort desc` or `-sdesc`. Leading dots are ignored during sorting.\n\n### Alignment\n\nA lot of code related files start with a leading dot `.` for no valid reason. `pls` by default\n\n- moves their name left by one character to line up the actual alphabets\n- dims their leading dot\n\nIf you don\'t like this, you can set `--no-align` to turn off all this behaviour in one swoop.\n\n### Details\n\nWhen you need more infomation about your files, pass the `--details` flag. This expands the list into a table, with\n\n- permissions\n- owner name\n- size\n\nadded to the output. The permissions are presented as `rwx` triplets. The size is presented in binary compound-units (the ones with the "i" like "*iB"). You can switch to decimal units by passing `--units decimal` or `-udecimal`. This flag has no effect unless the `--detail` flag is passed too.\n',
    'author': 'Dhruv Bhanushali',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dhruvkb/pls',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



class ConfigException(Exception):
    """
    These exceptions occur as a result of invalid configuration and are largely
    internal. The user should not experience these exceptions unless they are
    modifying the configs.

    The output of these exceptions must be detailed and developer-oriented.
    """

    pass


class ExecException(Exception):
    """
    These exceptions occur as a result of bad input and cause the application to
    terminate. These exceptions will be seen by the user.

    The output of these exceptions must be succinct and helpful.
    """

    pass
